from aumbry.contract import AbstractSource
from aumbry.errors import UnknownSourceError
from pike.discovery import py
from pike.manager import PikeManager


def find_plugin(name, base_cls, search_paths):
    with PikeManager(search_paths) as mgr:
        handlers = mgr.get_all_inherited_classes(base_cls)

        for handler in handlers:
            if handler.extras_name == name:
                return handler


def find_source(name, search_paths=None):
    if not search_paths:
        search_paths = py.get_module_by_name('aumbry.sources').__path__

    return find_plugin(name, AbstractSource, search_paths)


def load(source_name, config_class, options=None, search_paths=None):
    """Loads a configration from a source into the specified Config type

    Args:
        source_name (str): The name of the desired source.
        config_class (AumbryConfig): The resulting class of configuration you
            wish to deserialize the data into.
        options (dict, optional): The options used by the source handler. The
            keys are determined by each source handler. Refer to your source
            handler documentation on what options are available.
        search_paths (list, optional): A list paths for custom source handlers

    Returns:
        An instance of the passed in config_class
    """
    source_cls = find_source(source_name, search_paths)
    if not source_cls:
        raise UnknownSourceError(source_name)

    source = source_cls(options)
    handler = config_class.__handler__()

    source.import_requirements()
    handler.import_requirements()

    cfg_data = source.fetch_config_data()
    return handler.deserialize(cfg_data, config_class)


def save(source_name, config_inst, options=None, search_paths=None):
    """Loads a configration from a source into the specified Config type

    Args:
        source_name (str): The name of the desired source.
        config_inst (AumbryConfig): The instance of a configuration class
            wish save.
        options (dict, optional): The options used by the source handler. The
            keys are determined by each source handler. Refer to your source
            handler documentation on what options are available.
        search_paths (list, optional): A list paths for custom source handlers
    """
    source_cls = find_source(source_name, search_paths)
    if not source_cls:
        raise UnknownSourceError(source_name)

    source = source_cls(options)
    handler = config_inst.__handler__()

    source.import_requirements()
    handler.import_requirements()

    data = handler.serialize(config_inst)
    return source.save_config_data(data)



from alchemize.transmute import JsonTransmuter

from aumbry.contract import AbstractHandler, AumbryConfig


class JsonHandler(AbstractHandler):
    extras_name = 'json'

    @property
    def imports(self):
        return ['json']

    def serialize(self, config):
        return JsonTransmuter.transmute_to(config)

    def deserialize(self, raw_config, config_cls):
        return JsonTransmuter.transmute_from(
            raw_config.decode('utf-8'),
            config_cls
        )


class JsonConfig(AumbryConfig):
    """ A type of AumbryConfig for JSON Configurations."""
    __handler__ = JsonHandler



from alchemize.transmute import JsonTransmuter

from aumbry.contract import AumbryConfig
from aumbry.formats import js


class YamlHandler(js.JsonHandler):
    extras_name = 'yaml'

    @property
    def imports(self):
        return ['yaml']

    def serialize(self, config):
        import yaml

        config_dict = JsonTransmuter.transmute_to(config, to_string=False)
        return yaml.dump(config_dict)

    def deserialize(self, raw_config, config_cls):
        import yaml

        config_dict = yaml.load(raw_config)
        return JsonTransmuter.transmute_from(config_dict, config_cls)


class YamlConfig(AumbryConfig):
    """ A type of AumbryConfig for Yaml Configurations."""
    __handler__ = YamlHandler



from aumbry.contract import AbstractSource
from aumbry.sources import SourceTypes
from aumbry.utils.file import load_file, save_file


class FileSource(AbstractSource):
    extras_name = SourceTypes.file

    @property
    def imports(self):
        return []

    @property
    def environment_var_prefix(self):
        return 'CONFIG_FILE'

    def fetch_config_data(self):
        path = self.vars['CONFIG_FILE_PATH']
        return load_file(path)

    def save_config_data(self, data):
        path = self.vars['CONFIG_FILE_PATH']
        return save_file(path, data)



import os
from aumbry.errors import LoadError, SaveError


def load_file(filename):
    if not os.path.exists(filename):
        raise LoadError('Path {}, doesn\'t exist!'.format(filename))

    try:
        with open(filename, 'rb') as fp:
            return fp.read()
    except Exception as e:
        raise LoadError(str(e))


def save_file(filename, data):
    try:
        with open(filename, 'wb') as fp:
            return fp.write(data.encode('utf-8'))
    except Exception as e:
        raise SaveError(str(e))



"""
Attributes:
    FILE (str): Alias of SourceTypes.file
    CONSUL (str): Alias of SourceTypes.consul
"""
from .loader import load, save # NOQA
from .formats.js import JsonConfig # NOQA
from .formats.yml import YamlConfig # NOQA
from .sources import SourceTypes

from alchemize.mapping import Attr # NOQA

FILE = SourceTypes.file
CONSUL = SourceTypes.consul


__all__ = [
    'Attr',
    'load',
    'save',
    'JsonConfig',
    'YamlConfig',
    'SourceTypes',
]



class AumbryError(Exception):
    def __init__(self, message):
        self.message = message
        super(AumbryError, self).__init__(message)


class LoadError(AumbryError):
    pass


class SaveError(AumbryError):
    pass


class ParsingError(AumbryError):
    pass


class DependencyError(AumbryError):
    def __init__(self, extras_name):
        msg = (
            'Dependencies unavailable: run "pip install aumbry[{}]" to '
            'acquire to appropriate dependencies.'
        ).format(
            extras_name
        )
        super(DependencyError, self).__init__(msg)


class UnknownSourceError(AumbryError):
    def __init__(self, name):
        super(UnknownSourceError, self).__init__(
            'Couldn\'t find a source with the name: {}'.format(name)
        )



import time
import base64
from six.moves.urllib.parse import urljoin

from aumbry.contract import AbstractSource
from aumbry.errors import LoadError
from aumbry.sources import SourceTypes


class ConsulSource(AbstractSource):
    extras_name = SourceTypes.consul

    @property
    def imports(self):
        return ['requests']

    @property
    def environment_var_prefix(self):
        return 'CONSUL'

    def fetch_config_data(self):
        import requests

        consul_uri = self.vars['CONSUL_URI']
        consul_key = self.vars['CONSUL_KEY']
        timeout = self.vars.get('CONSUL_TIMEOUT', 10)
        retries = self.vars.get('CONSUL_RETRY_MAX', 1)
        interval = self.vars.get('CONSUL_RETRY_INTERVAL', 10)

        full_uri = urljoin(consul_uri, '/v1/kv/{}'.format(consul_key))

        tries = 0
        while tries <= retries:
            tries += 1
            resp = requests.get(full_uri, timeout=timeout)
            if resp.status_code == 200:
                data = resp.json()[0].get('Value')
                return base64.b64decode(data)
            elif resp.status_code == 404:
                raise LoadError(
                    'Consul returned 404 when fetching {}'.format(full_uri)
                )

            time.sleep(interval)

        # If we've made it this far... boom!
        msg = ('Hit max retry attempts. Consul returned {} when '
               'fetching {}').format(resp.status_code, full_uri)
        raise LoadError(msg)

    def save_config_data(self, data):
        raise NotImplementedError()



import abc
import os
import importlib
import six

from alchemize.mapping import JsonMappedModel
from aumbry.errors import DependencyError


class PluginBase(object):
    @abc.abstractproperty
    def extras_name(self):
        """ The name of the extras package requirement. """

    @abc.abstractproperty
    def imports(self):
        """ List of modules to import. """

    def import_requirements(self):
        def _import(req, pkg=None):
            try:
                importlib.import_module(req, pkg)
            except:
                raise DependencyError(self.extras_name)

        for req in self.imports:
            if isinstance(req, six.text_type) or isinstance(req, str):
                _import(req)
            else:
                name, package = req
                _import(name, package)


@six.add_metaclass(abc.ABCMeta)
class AbstractSource(PluginBase):
    def __init__(self, options=None):
        self.options = options or {}

        self.vars = {}
        self.vars.update(self.environmental_vars)
        self.vars.update(self.options)

    @abc.abstractmethod
    def fetch_config_data(self):
        """ Retrieves configuration data from pre-configured location. """

    @abc.abstractproperty
    def environment_var_prefix(self):
        """ The prefix of all environmental variables to fetch. """

    @property
    def environmental_vars(self):
        keys_to_fetch = [
            key for key in os.environ.keys()
            if key.startswith(self.environment_var_prefix)
        ]

        return {key: os.environ[key] for key in keys_to_fetch}


@six.add_metaclass(abc.ABCMeta)
class AbstractHandler(PluginBase):
    @abc.abstractmethod
    def deserialize(self, raw_config, config_cls):
        """ Method to handle deserialization to a Config object. """

    @abc.abstractmethod
    def serialize(self, config):
        """ Method to handle serialization to a string. """


class AumbryConfig(JsonMappedModel):
    __handler__ = None
